from numbers import Number

from busypie.durations import SECOND


def time_value_operator(value, unit=SECOND, visitor=None):
    _validate_time_and_unit(value, unit)
    return visitor(value * unit)


def _validate_time_and_unit(value, unit):
    _validate_positive_number(value, 'Time value of {} is not allowed')
    _validate_positive_number(unit, 'Unit value of {} is not allowed')


def _validate_positive_number(value, message):
    if value is None or not isinstance(value, Number) or value < 0:
        raise ValueError(message.format(value))


from copy import deepcopy
from functools import partial

from busypie.awaiter import ConditionAwaiter
from busypie.durations import SECOND, ONE_HUNDRED_MILLISECONDS
from busypie.time import time_value_operator

DEFAULT_MAX_WAIT_TIME = 10 * SECOND
DEFAULT_POLL_INTERVAL = ONE_HUNDRED_MILLISECONDS
DEFAULT_POLL_DELAY = ONE_HUNDRED_MILLISECONDS


class ConditionBuilder:

    def __init__(self, condition=None):
        self._condition = Condition() if condition is None else condition
        self._create_time_based_functions()

    def _create_time_based_functions(self):
        self.at_most = self._time_property_func_for('wait_time_in_secs')
        self.wait_at_most = self.at_most
        self.poll_delay = self._time_property_func_for('poll_delay')
        self.poll_interval = self._time_property_func_for('poll_interval')

    def _time_property_func_for(self, name):
        return partial(time_value_operator, visitor=partial(self._time_property, name=name))

    def _time_property(self, value, name):
        setattr(self._condition, name, value)
        return self._new_builder_with_cloned_condition()

    def _new_builder_with_cloned_condition(self):
        return ConditionBuilder(deepcopy(self._condition))

    def ignore_exceptions(self, *excludes):
        self._condition.ignored_exceptions = excludes
        return self._new_builder_with_cloned_condition()

    def wait(self):
        return self._new_builder_with_cloned_condition()

    def until(self, func):
        ConditionAwaiter(
            condition=self._condition,
            func_checker=lambda f: f()).wait_for(func)

    def during(self, func):
        ConditionAwaiter(
            condition=self._condition,
            func_checker=lambda f: not f()).wait_for(func)

    def __eq__(self, other):
        if not isinstance(other, ConditionBuilder):
            return False
        return self._condition == other._condition


class Condition:
    wait_time_in_secs = DEFAULT_MAX_WAIT_TIME
    ignored_exceptions = None
    poll_interval = DEFAULT_POLL_INTERVAL
    poll_delay = DEFAULT_POLL_DELAY

    def __eq__(self, other):
        if not isinstance(other, Condition):
            return False
        return \
            self.wait_time_in_secs == other.wait_time_in_secs and \
            self.ignored_exceptions == other.ignored_exceptions and \
            self.poll_interval == other.poll_interval and \
            self.poll_delay == other.poll_delay


set_default_timeout = partial(
    time_value_operator,
    visitor=partial(setattr, Condition, 'wait_time_in_secs'))


def reset_defaults():
    Condition.wait_time_in_secs = DEFAULT_MAX_WAIT_TIME



import busypie.condition
from busypie.condition import ConditionBuilder
from busypie.durations import SECOND


def wait():
    return ConditionBuilder()


given = wait


def wait_at_most(value, unit=SECOND):
    return wait().at_most(value, unit)


set_default_timeout = busypie.condition.set_default_timeout

reset_defaults = busypie.condition.reset_defaults



import time
from time import sleep

import busypie


class ConditionAwaiter:
    def __init__(self, condition, func_checker):
        self._condition = condition
        self._func_check = func_checker
        self._validate_condition()

    def _validate_condition(self):
        if self._condition.poll_delay > self._condition.wait_time_in_secs:
            raise ValueError('Poll delay should be shorter than maximum wait constraint')

    def wait_for(self, func):
        start_time = time.time()
        sleep(self._condition.poll_delay)
        while True:
            try:
                if self._func_check(func):
                    break
            except Exception as e:
                self._raise_exception_if_not_ignored(e)
            self._validate_wait_constraint(func.__name__, start_time)
            sleep(self._condition.poll_interval)

    def _raise_exception_if_not_ignored(self, e):
        ignored_exceptions = self._condition.ignored_exceptions
        if ignored_exceptions is None or \
                (ignored_exceptions and e.__class__ not in ignored_exceptions):
            raise e

    def _validate_wait_constraint(self, func_name, start_time):
        if (time.time() - start_time) > self._condition.wait_time_in_secs:
            raise busypie.ConditionTimeoutError("Failed to meet condition of {} within {} seconds"
                                                .format(func_name, self._condition.wait_time_in_secs))


class ConditionTimeoutError(Exception):
    pass
